import contextlib
import functools
import inspect
import types
import typing


class Deferred:
    __slots__: typing.List = ["deferred"]
    deferred: typing.List

    def __init__(self):
        self.deferred = list()

    def add(self, f: typing.Callable, *args: typing.Any, **kwargs: typing.Any) -> None:
        """
        Add a deferred function call.

        :param f: Function to be called
        :param args: Arguments to pass into the function
        :param kwargs: Keyword arguments to pass into the function.
        """
        self.deferred.append(lambda: f(*args, **kwargs))

    def close(self):
        """
        Loop through all deferred functions and call them.

        Method is named close to support contextlib.closing.
        """
        for f in self.deferred:  # ensure deferrals ran synchronously.
            f()


def with_defer(f: typing.Callable) -> typing.Callable:
    """
    Decorate a function to include support for function deferrals.

    :param f: Function to decorate, currently unsupported on classes.
    """

    @functools.wraps(f)
    def wrapper(*args: typing.Any, **kwargs: typing.Any):
        __deferred__: Deferred = Deferred()
        with contextlib.closing(__deferred__):
            return f(*args, **kwargs)

    return wrapper if isinstance(f, types.FunctionType) else f


def defer(f: typing.Callable, *args: typing.Any, **kwargs: typing.Any) -> None:
    """
    Defer a function call until the very end.

    :param f: Function to be called
    :param args: Arguments to pass into the function
    :param kwargs: Keyword arguments to pass into the function.
    """

    # loop the local variables starting from 2 functions up, adding the deferred function
    # if it can find a local with the name `__deferred__` and of type defer.Deferred
    for local in inspect.stack(context=0)[2:]:
        if isinstance(local[0].f_locals.get("__deferred__"), Deferred):
            deferred: Deferred = local[0].f_locals["__deferred__"]
            deferred.add(f, *args, **kwargs)
            break
    else:
        raise ValueError("Function not decorated with `defer.with_defer`")



#!/usr/bin/env python
# Learn more: https://github.com/kennethreitz/setup.py
import codecs
import os
import sys

import setuptools
import setuptools.command.test

here = os.path.abspath(os.path.dirname(__file__))


class PyTest(setuptools.command.test.test):
    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        setuptools.command.test.test.initialize_options(self)
        try:
            from multiprocessing import cpu_count

            self.pytest_args = [
                "-n",
                str(cpu_count()),
                "--boxed",
                "--mypy",
                "--ignore=setup.py",
            ]
        except (ImportError, NotImplementedError):
            self.pytest_args = ["-n", "1", "--boxed", "--mypy", "--ignore=setup.py"]

    def finalize_options(self):
        setuptools.command.test.test.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


# 'setup.py publish' shortcut.
if sys.argv[-1] == "publish":
    os.system("python -m mypy.stubgen -m defer -o .")
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/*")
    sys.exit()

packages = ["defer"]

requires = []
test_requirements = [
    "pytest-cov>=2.6.1",
    "pytest>=4.3.1",
    "pytest-xdist>=1.27.0",
    "pytest-cov>=2.6.1",
]

with codecs.open("README.md", "r", "utf-8") as f:
    readme = f.read()
with codecs.open("HISTORY.md", "r", "utf-8") as f:
    history = f.read()

__title__: str = "py-defer"
__description__: str = "Golang-esque defer functionality"
__url__: str = "http://github.com/naphta/with-defer"
__version__: str = "1.1.2"
__author__: str = "Jake Hill"
__author_email__: str = "jake@naphta.uk"
__license__: str = "MIT"
__copyright__: str = "Copyright 2018 Jake Hill"


setuptools.setup(
    name=__title__,
    version=__version__,
    description=__description__,
    long_description=readme,
    long_description_content_type="text/markdown",
    author=__author__,
    author_email=__author_email__,
    url=__url__,
    packages=packages,
    package_data={"": ["LICENSE"], "defer": ["*.pyi"]},
    package_dir={"defer": "defer"},
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=requires,
    license=__license__,
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    cmdclass={"test": PyTest},
    tests_require=test_requirements,
)
